
package org.mule.modules.salesforce.config;

import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.lang.reflect.WildcardType;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;
import com.sforce.ws.transport.SoapConnection.SessionTimedOutException;
import org.mule.api.MessagingException;
import org.mule.api.MuleContext;
import org.mule.api.MuleEvent;
import org.mule.api.MuleException;
import org.mule.api.MuleMessage;
import org.mule.api.construct.FlowConstruct;
import org.mule.api.construct.FlowConstructAware;
import org.mule.api.context.MuleContextAware;
import org.mule.api.expression.ExpressionManager;
import org.mule.api.lifecycle.Disposable;
import org.mule.api.lifecycle.Initialisable;
import org.mule.api.lifecycle.InitialisationException;
import org.mule.api.lifecycle.Startable;
import org.mule.api.lifecycle.Stoppable;
import org.mule.api.processor.MessageProcessor;
import org.mule.api.registry.RegistrationException;
import org.mule.api.transformer.DataType;
import org.mule.api.transformer.Transformer;
import org.mule.api.transformer.TransformerException;
import org.mule.config.i18n.CoreMessages;
import org.mule.config.i18n.MessageFactory;
import org.mule.modules.salesforce.SalesforceModule;
import org.mule.transformer.TransformerTemplate;
import org.mule.transformer.types.DataTypeFactory;
import org.mule.transport.NullPayload;
import org.mule.util.TemplateParser;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


/**
 * QueryMessageProcessor invokes the {@link org.mule.modules.salesforce.SalesforceModule#query(java.lang.String)} method in {@link SalesforceModule }. For each argument there is a field in this processor to match it.  Before invoking the actual method the processor will evaluate and transform where possible to the expected argument type.
 * 
 */
public class QueryMessageProcessor
    implements FlowConstructAware, MuleContextAware, Disposable, Initialisable, Startable, Stoppable, MessageProcessor
{

    private Object query;
    private String _queryType;
    private Object username;
    private String _usernameType;
    private Object password;
    private String _passwordType;
    private Object securityToken;
    private String _securityTokenType;
    private static Logger logger = LoggerFactory.getLogger(QueryMessageProcessor.class);
    /**
     * Module object
     * 
     */
    private Object moduleObject;
    /**
     * Mule Context
     * 
     */
    private MuleContext muleContext;
    /**
     * Mule Expression Manager
     * 
     */
    private ExpressionManager expressionManager;
    /**
     * Mule Pattern Info
     * 
     */
    private TemplateParser.PatternInfo patternInfo;
    /**
     * Flow construct
     * 
     */
    private FlowConstruct flowConstruct;
    /**
     * Variable used to track how many retries we have attempted on this message processor
     * 
     */
    private AtomicInteger retryCount;
    /**
     * Maximum number of retries that can be attempted.
     * 
     */
    private int retryMax;

    /**
     * Obtains the expression manager from the Mule context and initialises the connector. If a target object  has not been set already it will search the Mule registry for a default one.
     * 
     * @throws InitialisationException
     */
    public void initialise()
        throws InitialisationException
    {
        retryCount = new AtomicInteger();
        expressionManager = muleContext.getExpressionManager();
        patternInfo = TemplateParser.createMuleStyleParser().getStyle();
        if (moduleObject == null) {
            try {
                moduleObject = muleContext.getRegistry().lookupObject(SalesforceModuleConnectionManager.class);
                if (moduleObject == null) {
                    throw new InitialisationException(MessageFactory.createStaticMessage("Cannot find object"), this);
                }
            } catch (RegistrationException e) {
                throw new InitialisationException(CoreMessages.initialisationFailure("org.mule.modules.salesforce.config.SalesforceModuleConnectionManager"), e, this);
            }
        }
    }

    public void start()
        throws MuleException
    {
    }

    public void stop()
        throws MuleException
    {
    }

    public void dispose() {
    }

    /**
     * Set the Mule context
     * 
     * @param context Mule context to set
     */
    public void setMuleContext(MuleContext context) {
        this.muleContext = context;
    }

    /**
     * Sets flow construct
     * 
     * @param flowConstruct Flow construct to set
     */
    public void setFlowConstruct(FlowConstruct flowConstruct) {
        this.flowConstruct = flowConstruct;
    }

    /**
     * Sets the instance of the object under which the processor will execute
     * 
     * @param moduleObject Instace of the module
     */
    public void setModuleObject(Object moduleObject) {
        this.moduleObject = moduleObject;
    }

    /**
     * Sets retryMax
     * 
     * @param value Value to set
     */
    public void setRetryMax(int value) {
        this.retryMax = value;
    }

    /**
     * Sets query
     * 
     * @param value Value to set
     */
    public void setQuery(Object value) {
        this.query = value;
    }

    /**
     * Sets username
     * 
     * @param value Value to set
     */
    public void setUsername(Object value) {
        this.username = value;
    }

    /**
     * Sets securityToken
     * 
     * @param value Value to set
     */
    public void setSecurityToken(Object value) {
        this.securityToken = value;
    }

    /**
     * Sets password
     * 
     * @param value Value to set
     */
    public void setPassword(Object value) {
        this.password = value;
    }

    /**
     * Get all superclasses and interfaces recursively.
     * 
     * @param classes List of classes to which to add all found super classes and interfaces.
     * @param clazz   The class to start the search with.
     */
    private void computeClassHierarchy(Class clazz, List classes) {
        for (Class current = clazz; (current!= null); current = current.getSuperclass()) {
            if (classes.contains(current)) {
                return ;
            }
            classes.add(current);
            for (Class currentInterface: current.getInterfaces()) {
                computeClassHierarchy(currentInterface, classes);
            }
        }
    }

    /**
     * Checks whether the specified class parameter is an instance of {@link List }
     * 
     * @param clazz <code>Class</code> to check.
     * @return
     */
    private boolean isListClass(Class clazz) {
        List<Class> classes = new ArrayList<Class>();
        computeClassHierarchy(clazz, classes);
        return classes.contains(List.class);
    }

    /**
     * Checks whether the specified class parameter is an instance of {@link Map }
     * 
     * @param clazz <code>Class</code> to check.
     * @return
     */
    private boolean isMapClass(Class clazz) {
        List<Class> classes = new ArrayList<Class>();
        computeClassHierarchy(clazz, classes);
        return classes.contains(Map.class);
    }

    private boolean isList(Type type) {
        if ((type instanceof Class)&&isListClass(((Class) type))) {
            return true;
        }
        if (type instanceof ParameterizedType) {
            return isList(((ParameterizedType) type).getRawType());
        }
        if (type instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) type).getUpperBounds();
            return ((upperBounds.length!= 0)&&isList(upperBounds[ 0 ]));
        }
        return false;
    }

    private boolean isMap(Type type) {
        if ((type instanceof Class)&&isMapClass(((Class) type))) {
            return true;
        }
        if (type instanceof ParameterizedType) {
            return isMap(((ParameterizedType) type).getRawType());
        }
        if (type instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) type).getUpperBounds();
            return ((upperBounds.length!= 0)&&isMap(upperBounds[ 0 ]));
        }
        return false;
    }

    private boolean isAssignableFrom(Type expectedType, Class clazz) {
        if (expectedType instanceof Class) {
            return ((Class) expectedType).isAssignableFrom(clazz);
        }
        if (expectedType instanceof ParameterizedType) {
            return isAssignableFrom(((ParameterizedType) expectedType).getRawType(), clazz);
        }
        if (expectedType instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) expectedType).getUpperBounds();
            if (upperBounds.length!= 0) {
                return isAssignableFrom(upperBounds[ 0 ], clazz);
            }
        }
        return false;
    }

    private Object evaluate(MuleMessage muleMessage, Object source) {
        if (source instanceof String) {
            String stringSource = ((String) source);
            if (stringSource.startsWith(patternInfo.getPrefix())&&stringSource.endsWith(patternInfo.getSuffix())) {
                return expressionManager.evaluate(stringSource, muleMessage);
            } else {
                return expressionManager.parse(stringSource, muleMessage);
            }
        }
        return source;
    }

    private Object evaluateAndTransform(MuleMessage muleMessage, Type expectedType, Object source)
        throws TransformerException
    {
        if (source == null) {
            return source;
        }
        Object target = null;
        if (isList(source.getClass())) {
            if (isList(expectedType)) {
                List newList = new ArrayList();
                Type valueType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 0 ];
                ListIterator iterator = ((List) source).listIterator();
                while (iterator.hasNext()) {
                    Object subTarget = iterator.next();
                    newList.add(evaluateAndTransform(muleMessage, valueType, subTarget));
                }
                target = newList;
            } else {
                target = source;
            }
        } else {
            if (isMap(source.getClass())) {
                if (isMap(expectedType)) {
                    Type keyType = Object.class;
                    Type valueType = Object.class;
                    if (expectedType instanceof ParameterizedType) {
                        keyType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 0 ];
                        valueType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 1 ];
                    }
                    Map map = ((Map) source);
                    Map newMap = new HashMap();
                    for (Object entryObj: map.entrySet()) {
                        {
                            Map.Entry entry = ((Map.Entry) entryObj);
                            Object newKey = evaluateAndTransform(muleMessage, keyType, entry.getKey());
                            Object newValue = evaluateAndTransform(muleMessage, valueType, entry.getValue());
                            newMap.put(newKey, newValue);
                        }
                    }
                    target = newMap;
                } else {
                    target = source;
                }
            } else {
                target = evaluate(muleMessage, source);
            }
        }
        if ((target!= null)&&(!isAssignableFrom(expectedType, target.getClass()))) {
            DataType sourceDataType = DataTypeFactory.create(target.getClass());
            DataType targetDataType = DataTypeFactory.create(((Class) expectedType));
            Transformer t = muleContext.getRegistry().lookupTransformer(sourceDataType, targetDataType);
            return t.transform(target);
        } else {
            return target;
        }
    }

    /**
     * Invokes the MessageProcessor.
     * 
     * @param event MuleEvent to be processed
     * @throws MuleException
     */
    public MuleEvent process(MuleEvent event)
        throws MuleException
    {
        MuleMessage muleMessage = event.getMessage();
        SalesforceModuleConnectionManager castedModuleObject = null;
        if (moduleObject instanceof String) {
            castedModuleObject = ((SalesforceModuleConnectionManager) muleContext.getRegistry().lookupObject(((String) moduleObject)));
            if (castedModuleObject == null) {
                throw new MessagingException(CoreMessages.failedToCreate("query"), event, new RuntimeException("Cannot find the configuration specified by the config-ref attribute."));
            }
        } else {
            castedModuleObject = ((SalesforceModuleConnectionManager) moduleObject);
        }
        String transformedUsername = null;
        String transformedPassword = null;
        String transformedSecurityToken = null;
        SalesforceModuleLifecycleAdapter connection = null;
        try {
            if (username!= null) {
                transformedUsername = ((String) evaluateAndTransform(muleMessage, QueryMessageProcessor.class.getDeclaredField("_usernameType").getGenericType(), username));
            } else {
                if (castedModuleObject.getUsername() == null) {
                    throw new MessagingException(CoreMessages.failedToCreate("query"), event, new RuntimeException("You must provide a username at the config or the message processor level."));
                }
                transformedUsername = ((String) evaluateAndTransform(muleMessage, QueryMessageProcessor.class.getDeclaredField("_usernameType").getGenericType(), castedModuleObject.getUsername()));
            }
            if (password!= null) {
                transformedPassword = ((String) evaluateAndTransform(muleMessage, QueryMessageProcessor.class.getDeclaredField("_passwordType").getGenericType(), password));
            } else {
                if (castedModuleObject.getPassword() == null) {
                    throw new MessagingException(CoreMessages.failedToCreate("query"), event, new RuntimeException("You must provide a password at the config or the message processor level."));
                }
                transformedPassword = ((String) evaluateAndTransform(muleMessage, QueryMessageProcessor.class.getDeclaredField("_passwordType").getGenericType(), castedModuleObject.getPassword()));
            }
            if (securityToken!= null) {
                transformedSecurityToken = ((String) evaluateAndTransform(muleMessage, QueryMessageProcessor.class.getDeclaredField("_securityTokenType").getGenericType(), securityToken));
            } else {
                if (castedModuleObject.getSecurityToken() == null) {
                    throw new MessagingException(CoreMessages.failedToCreate("query"), event, new RuntimeException("You must provide a securityToken at the config or the message processor level."));
                }
                transformedSecurityToken = ((String) evaluateAndTransform(muleMessage, QueryMessageProcessor.class.getDeclaredField("_securityTokenType").getGenericType(), castedModuleObject.getSecurityToken()));
            }
            String transformedQuery = ((String) evaluateAndTransform(muleMessage, QueryMessageProcessor.class.getDeclaredField("_queryType").getGenericType(), query));
            if (logger.isDebugEnabled()) {
                StringBuffer messageStringBuffer = new StringBuffer();
                messageStringBuffer.append("Attempting to acquire a connection using ");
                messageStringBuffer.append("[username = ");
                messageStringBuffer.append(transformedUsername);
                messageStringBuffer.append("] ");
                messageStringBuffer.append("[securityToken = ");
                messageStringBuffer.append(transformedSecurityToken);
                messageStringBuffer.append("] ");
                messageStringBuffer.append("[password = ");
                messageStringBuffer.append(transformedPassword);
                messageStringBuffer.append("] ");
                logger.debug(messageStringBuffer.toString());
            }
            connection = castedModuleObject.acquireConnection(new SalesforceModuleConnectionManager.ConnectionParameters(transformedUsername, transformedSecurityToken, transformedPassword));
            if (connection == null) {
                throw new MessagingException(CoreMessages.failedToCreate("query"), event, new RuntimeException("Cannot create connection"));
            } else {
                if (logger.isDebugEnabled()) {
                    StringBuffer messageStringBuffer = new StringBuffer();
                    messageStringBuffer.append("Connection has been acquired with ");
                    messageStringBuffer.append("[id = ");
                    messageStringBuffer.append(connection.getSessionId());
                    messageStringBuffer.append("] ");
                    logger.debug(messageStringBuffer.toString());
                }
            }
            retryCount.getAndIncrement();
            Object resultPayload;
            resultPayload = connection.query(transformedQuery);
            TransformerTemplate.OverwitePayloadCallback overwritePayloadCallback = null;
            if (resultPayload == null) {
                overwritePayloadCallback = new TransformerTemplate.OverwitePayloadCallback(NullPayload.getInstance());
            } else {
                overwritePayloadCallback = new TransformerTemplate.OverwitePayloadCallback(resultPayload);
            }
            List<Transformer> transformerList;
            transformerList = new ArrayList<Transformer>();
            transformerList.add(new TransformerTemplate(overwritePayloadCallback));
            event.getMessage().applyTransformers(event, transformerList);
            retryCount.set(0);
            return event;
        } catch (SessionTimedOutException invalidConnection) {
            if (logger.isDebugEnabled()) {
                StringBuffer messageStringBuffer = new StringBuffer();
                messageStringBuffer.append("An exception (");
                messageStringBuffer.append("com.sforce.ws.transport.SoapConnection.SessionTimedOutException");
                messageStringBuffer.append(") has been thrown while executing ");
                messageStringBuffer.append("query");
                messageStringBuffer.append(". Destroying the connection with [id = ");
                messageStringBuffer.append(connection.getSessionId());
                messageStringBuffer.append("].");
                logger.debug(messageStringBuffer.toString());
            }
            try {
                castedModuleObject.destroyConnection(new SalesforceModuleConnectionManager.ConnectionParameters(transformedUsername, transformedSecurityToken, transformedPassword), connection);
                connection = null;
            } catch (Exception e) {
                logger.error(e.getMessage(), e);
            }
            if (retryCount.get()<= retryMax) {
                if (logger.isDebugEnabled()) {
                    StringBuffer messageStringBuffer = new StringBuffer();
                    messageStringBuffer.append("Forcing a retry [time=");
                    messageStringBuffer.append(retryCount);
                    messageStringBuffer.append(" out of  ");
                    messageStringBuffer.append(retryMax);
                    messageStringBuffer.append("].");
                    logger.debug(messageStringBuffer.toString());
                }
                return process(event);
            }
            throw new MessagingException(CoreMessages.failedToInvoke("query"), event, invalidConnection);
        } catch (Exception e) {
            throw new MessagingException(CoreMessages.failedToInvoke("query"), event, e);
        } finally {
            try {
                if (connection!= null) {
                    if (logger.isDebugEnabled()) {
                        StringBuffer messageStringBuffer = new StringBuffer();
                        messageStringBuffer.append("Releasing the connection back into the pool [id=");
                        messageStringBuffer.append(connection.getSessionId());
                        messageStringBuffer.append("].");
                        logger.debug(messageStringBuffer.toString());
                    }
                    castedModuleObject.releaseConnection(new SalesforceModuleConnectionManager.ConnectionParameters(transformedUsername, transformedSecurityToken, transformedPassword), connection);
                }
            } catch (Exception e) {
                throw new MessagingException(CoreMessages.failedToInvoke("query"), event, e);
            }
        }
    }

}
